from typing import Sequence
from .base import BaseNeuron


def input_transform(pops: Sequence):
  """Select BaseNeurons which locate on current process and return its input.

  Parameters
  ----------
  pops : Sequence

  Returns
  -------
  The inputs for the target DynamicalSystem. It should be the format
    of `[(target, value, [type, operation])]
  """
  if len(pops) > 0 and not isinstance(pops[0], (list, tuple)):
    pops = [pops]
  input_trans = []
  for pop in pops:
    try:
      tmp = pop[0].input
      input_trans.append((tmp, pop[1]) + pop[2:])
    except Exception as e:
      continue
  return input_trans


def monitor_transform(pops: Sequence[BaseNeuron], attr: str = 'spike'):
  """generate monitor parameters dictionary

  Parameters
  ----------
  pops : Sequence[BaseNeuron]

  attr : str, optional
      attribute to monitor, by default 'spike'

  Returns
  -------
  {attribute : BaseNeuron.attribute}
  """
  mon_var = {}
  for pop in pops:
    try:
      tmp = getattr(pop, attr)
      mon_var.update({attr: tmp})
    except Exception as e:
      continue
  return mon_var



from typing import Union, Callable, Optional

import brainpy.math as bm
from brainpy import dyn
from brainpy.initialize import Initializer, parameter
from brainpy.types import Array
from brainpy.dyn.base import TwoEndConn
import numpy as np
from bpl.core.base import DynamicalSystem
from mpi4py import MPI
import jax.numpy as jnp
import platform

if platform.system() != 'Windows':
  import mpi4jax


class RemoteSynapse(DynamicalSystem, dyn.base.TwoEndConn):
  """Remote synapse model in multi-device environment.
  This class is a wrapper which can run the given specific synapse model between 2 ranks.

  Parameters
  -------
  synapse_class: TwoEndConn
      Given specific synapse model class.
  param_dict: dict(python3.7 and follow-up version), OrderedDict
      A dict contains parameters which is needed by synapse_class.
  source_rank: int
      The rank id which pre neuron group is located in.
  target_rank: int
      The rank id which post neuron group is located in.
  comm: mpi4py.MPI.Intracomm
      Communicators object in mpi4py package.
  """

  def __init__(self, synapse_class, param_dict, source_rank, target_rank, comm=MPI.COMM_WORLD):
    # Make sure RemoteSynapse can work correctly.
    if not issubclass(synapse_class, TwoEndConn):
      raise Exception('synapse_class should be a subclass (not a instance of class) of brainpy.dyn.base.TwoEndConn.')
    count = 3
    for k, v in param_dict.items():
      if (count == 3 and k != 'pre') or (count == 2 and k != 'post') or (count == 1 and k != 'conn'):
        raise Exception('First 3 keys in param_dict must be pre post and conn in order.')
      count -= 1
    if count > 0:
      raise Exception(
        'At least 3 items should be given to param_dict and first 3 keys in param_dict must be pre post and conn in order.'
      )

    super(RemoteSynapse, self).__init__(param_dict['pre'], param_dict['post'], param_dict['conn'])

    # Create a new class because some class method will change in the following code.
    # Below code can prevent origin class from being affected.

    class sub_synapse_class(DynamicalSystem, synapse_class):
      pass

    self.synapse_class = sub_synapse_class

    self.comm = comm
    self.source_rank = source_rank
    self.target_rank = target_rank
    self.rank = self.comm.Get_rank()
    self.rank_pair = 'rank' + str(self.source_rank) + 'rank' + str(self.target_rank)
    # Replace some function to save place or make synapse work in muti-device enviornment
    self.synapse_class.register_delay = self.remote_register_delay

    if self.rank == source_rank:
      # Replace some function to save place or make synapse work in muti-device enviornment
      def source_rank_init_weights(*para, **dict):
        return None, None

      self.synapse_class.init_weights = source_rank_init_weights

      def variable_(*param, **dict):
        return None

      def odeint(*param, **dict):
        return None

      # Make sure the same neuron group only deliver its spike one time
      # during one step network simulation between this two ranks
      check_name = param_dict['pre'].name + self.rank_pair
      if check_name not in self.remote_synapse_mark:
        self.remote_synapse_mark.append(check_name)
        # send
        if platform.system() == 'Windows':
          self.comm.send(len(param_dict['pre'].spike), dest=target_rank, tag=0)
          self.comm.Send(param_dict['pre'].spike.to_numpy(), dest=target_rank, tag=1)
        else:
          token = mpi4jax.send(param_dict['pre'].spike.value, dest=target_rank, tag=0, comm=self.comm)
      # initialize
      self.synapse_instance = self.synapse_class(**param_dict)

    elif self.rank == target_rank:
      # Replace some function to save place or make synapse work in muti-device enviornment
      self.synapse_class.get_delay_data = self.remote_get_delay_data
      check_name = param_dict['pre'].name + self.rank_pair
      if check_name not in self.remote_synapse_mark:
        self.remote_synapse_mark.append(check_name)
        # receive
        if platform.system() == 'Windows':
          pre_len = self.comm.recv(source=source_rank, tag=0)
          pre_spike = np.empty(pre_len, dtype=np.bool_)
          self.comm.Recv(pre_spike, source=source_rank, tag=1)
        else:
          pre_spike, token = mpi4jax.recv(param_dict['pre'].spike.value, source=source_rank, tag=0, comm=self.comm)
        # self.pre.spike should catch the spike data sent from source rank.
        param_dict['pre'].spike = bm.Variable(pre_spike)
      # initialize
      self.synapse_instance = self.synapse_class(**param_dict)

  def remote_register_delay(
    self,
    identifier: str,
    delay_step: Optional[Union[int, Array, Callable, Initializer]],
    delay_target: bm.Variable,
    initial_delay_data: Union[Initializer, Callable, Array, float, int, bool] = None,
  ):
    """Register delay variable in multi-device enviornmrnt.
    """
    identifier = identifier + self.rank_pair
    # delay steps
    if delay_step is None:
      delay_type = 'none'
    elif isinstance(delay_step, (int, np.integer, jnp.integer)):
      delay_type = 'homo'
    elif isinstance(delay_step, (bm.ndarray, jnp.ndarray, np.ndarray)):
      if delay_step.size == 1 and delay_step.ndim == 0:
        delay_type = 'homo'
      else:
        delay_type = 'heter'
        delay_step = bm.asarray(delay_step)
    elif callable(delay_step):
      delay_step = parameter(delay_step, delay_target.shape, allow_none=False)
      delay_type = 'heter'
    else:
      raise ValueError(f'Unknown "delay_steps" type {type(delay_step)}, only support '
                       f'integer, array of integers, callable function, brainpy.init.Initializer.')
    if delay_type == 'heter':
      if delay_step.dtype not in [bm.int32, bm.int64]:
        raise ValueError('Only support delay steps of int32, int64. If your '
                         'provide delay time length, please divide the "dt" '
                         'then provide us the number of delay steps.')
      if delay_target.shape[0] != delay_step.shape[0]:
        raise ValueError(f'Shape is mismatched: {delay_target.shape[0]} != {delay_step.shape[0]}')
    if delay_type != 'none':
      max_delay_step = int(bm.max(delay_step))

    # delay target
    if delay_type != 'none':
      if not isinstance(delay_target, bm.Variable):
        raise ValueError(f'"delay_target" must be an instance of Variable, but we got {type(delay_target)}')

    # delay variable
    if delay_type != 'none':
      if identifier not in self.remote_global_delay_data:
        delay = bm.LengthDelay(delay_target, max_delay_step, initial_delay_data)
        self.remote_global_delay_data[identifier] = (delay, delay_target)
        self.local_delay_vars[identifier] = delay
      else:
        delay = self.remote_global_delay_data[identifier][0]
        if delay is None:
          delay = bm.LengthDelay(delay_target, max_delay_step, initial_delay_data)
          self.remote_global_delay_data[identifier] = (delay, delay_target)
          self.local_delay_vars[identifier] = delay
        elif delay.num_delay_step - 1 < max_delay_step:
          self.remote_global_delay_data[identifier][0].reset(delay_target, max_delay_step, initial_delay_data)
    else:
      if identifier not in self.remote_global_delay_data:
        delay = bm.LengthDelay(delay_target, 0)
        self.remote_global_delay_data[identifier] = (delay, delay_target)
        self.local_delay_vars[identifier] = delay
        # Above row of code does not exist in 'register_delay' method.
        # In multi-device environment, spike of source neuron group need be delivered during local_delay_vars being traversed when Network is updating.
    self.register_implicit_nodes(self.local_delay_vars)
    return delay_step

  def remote_get_delay_data(
    self,
    identifier: str,
    delay_step: Optional[Union[int, bm.JaxArray, jnp.DeviceArray]],
    *indices: Union[int, slice, bm.JaxArray, jnp.DeviceArray],
  ):
    """Get delay data according to the provided delay steps in multi-device enviornment.
    """
    identifier = identifier + self.rank_pair
    if delay_step is None:
      if bm.ndim(delay_step) == 0:
        return self.remote_global_delay_data[identifier][0](0, *indices)
      else:
        if len(indices) == 0:
          indices = (jnp.arange(delay_step.size),)
        return self.remote_global_delay_data[identifier][0](0, *indices)

    if identifier in self.remote_global_delay_data:
      if bm.ndim(delay_step) == 0:
        return self.remote_global_delay_data[identifier][0](delay_step, *indices)
      else:
        if len(indices) == 0:
          indices = (jnp.arange(delay_step.size),)
        return self.remote_global_delay_data[identifier][0](delay_step, *indices)

    elif identifier in self.local_delay_vars:
      if bm.ndim(delay_step) == 0:
        return self.local_delay_vars[identifier](delay_step)
      else:
        if len(indices) == 0:
          indices = (jnp.arange(delay_step.size),)
        return self.local_delay_vars[identifier](delay_step, *indices)

    else:
      raise ValueError(f'{identifier} is not defined in delay variables.')

  def update(self, tdi, pre_spike=None):
    if self.rank == self.target_rank:
      self.synapse_instance.update(tdi, pre_spike)
    # pass



import sys
import brainpy.dyn as dyn
from brainpy.modes import Mode, normal
from brainpy.types import Shape, Array
from brainpy.connect import TwoEndConnector
from typing import Union, Sequence, Dict
import jax.tree_util
from bpl.respa.res_manager import ResManager
import bpl
import copy

try:
  from mpi4py import MPI

  mpi_size = MPI.COMM_WORLD.Get_size()
  mpi_rank = MPI.COMM_WORLD.Get_rank()
except ImportError:
  mpi_size = 1
  mpi_rank = 0

pop_id = 0


def get_pop_id():
  global pop_id
  pop_id += 1
  return pop_id


def reset():
  global pop_id
  ResManager.clear()
  pop_id = 0


def set_platform(platform: str):
  """
  Changes platform to CPU, GPU, or TPU. This utility only takes
  effect at the beginning of your program.
  """
  from jax import config
  assert platform in ['cpu', 'gpu']
  config.update("jax_platform_name", platform)


class BaseNeuron:
  proxy_neurons = {}

  def __init__(self, shape: Shape, **kwargs):
    self.args = []
    self.kwargs = kwargs
    self.shape = shape
    # self.model_class = None
    self.lowref = None
    # process id
    self.pid = None
    # population id, autoincrement
    self.id = get_pop_id()
    self.neuron_start_id = 0
    ResManager.pops.append(self)
    ResManager.pops_by_id[self.id] = self

  def __hash__(self):
    return self.id

  def __eq__(self, other):
    return self.id == other.id

  def __getitem__(self, index: Union[slice, Sequence, Array]):
    return BaseNeuronSlice(self, index)

  def __getattr__(self, __name: str):
    return self.lowref.__getattribute__(__name)

  def build(self):  # TODO check current pid
    if self.lowref is not None:
      return self.lowref
    if not hasattr(self, 'model_class'):
      raise Exception("model_class should be assigned")
    self.lowref = self.model_class(self.shape, *self.args, **self.kwargs)
    return self.lowref


class BaseNeuronSlice:

  def __init__(self, target: Union[BaseNeuron, 'BaseNeuronSlice'], index: Union[slice, Sequence, Array]):
    if isinstance(target, BaseNeuron):
      self.target = target
      self.index = [index]
    elif isinstance(target, BaseNeuronSlice):
      self.target = target.target
      self.index = copy.deepcopy(target.index)
      self.index.append(index)
    else:
      raise ValueError("target should be a BaseNeuronSlice or a BaseNeuron")

  def build(self):
    if '_cache' not in self.__dict__:
      self._cache = self.target.build()
      for idx in self.index:
        self._cache = self._cache[idx]
    return self._cache

  def __getattr__(self, __name: str):
    if __name in self.target.__dict__:
      return self.target.__getattribute__(__name)
    else:
      return self.build().__getattribute__(__name)

  def __getitem__(self, index: Union[slice, Sequence, Array]):
    return BaseNeuronSlice(self, index)


class register():
  """decorator class to register user defined neuron model and synapse model to respa and base module

  **Decorator Examples**

    >>> import bpl
    >>> from brainpy.dyn import channels
    >>> @bpl.register()
    >>> class HH(bp.dyn.CondNeuGroup):
    >>>   def __init__(self, size):
    >>>     super(HH, self).__init__(size, )
    >>>     self.INa = channels.INa_TM1991(size, g_max=100., V_sh=-63.)
    >>>     self.IK = channels.IK_TM1991(size, g_max=30., V_sh=-63.)
    >>>     self.IL = channels.IL(size, E=-60., g_max=0.05)
    >>> pop = bpl.HH(100)

  Parameters
  ----------
  name : str = None
    The name of the class in the bpl, respa and base module
    User should use this class to create their model instance.
  model : str = None
    The type of the model
    If model is None the type will be inferred from the class.
  """

  def __init__(self, name: str = None, model: str = None):
    self.name = name
    self.model = model.lower() if model is not None else None

  def __call__(self, cls):
    if self.model is None and issubclass(cls, dyn.NeuGroup) or self.model == 'neuron':
      if self.name is None:
        name = cls.__name__
      else:
        name = self.name
      bases = (BaseNeuron,)
      attrs = {'__qualname__': name, 'model_class': cls}
    elif self.model is None and issubclass(cls, dyn.SynConn) or self.model == 'synapse':
      if self.name is None:
        name = cls.__name__
      else:
        name = self.name
      bases = (BaseSynapse,)
      attrs = {'__qualname__': name, 'model_class': cls, 'model_class_remote': None}
    else:
      raise RuntimeError(f'custom {self.model} is not supported')
    respa_type = type(name, bases, attrs)
    name_split = __name__.split('.')
    if len(name_split) > 1:
      setattr(sys.modules['.'.join(name_split[:-1])], name, respa_type)
    if len(name_split) > 2:
      setattr(sys.modules['.'.join(name_split[:-2])], name, respa_type)
    setattr(sys.modules[__name__], name, respa_type)
    return cls


class BaseSynapse:

  def __init__(self, pre: Union[BaseNeuron, BaseNeuronSlice], post: Union[BaseNeuron, BaseNeuronSlice],
               conn: Union[TwoEndConnector, Array, Dict[str, Array]], **kwargs):
    self.pre = pre
    self.post = post
    self.conn = conn
    self.args = []
    self.kwargs = kwargs
    # self.model_class = None
    self.lowref = None
    ResManager.syns.append(self)

  def __getattr__(self, __name: str):
    return self.lowref.__getattribute__(__name)

  def build(self):
    # assert self.pre.lowref is not None and self.post.lowref is not None
    if self.pre.pid != mpi_rank and self.post.pid != mpi_rank or self.lowref is not None:
      return self.lowref

    if not hasattr(self, 'model_class'):
      raise Exception("model_class should be assigned")

    pre_pid = self.pre.pid
    pre_shape = self.pre.shape
    if isinstance(self.pre, BaseNeuronSlice):
      pre_slice = self.pre.index
      if pre_pid == mpi_rank:
        pre = self.pre.build()
      else:
        pre = self.pre.target
    elif isinstance(self.pre, BaseNeuron):
      pre_slice = None
      if pre_pid == mpi_rank:
        pre = self.pre.build()
      else:
        pre = self.pre
    else:
      raise ValueError(type(self.pre))

    post_pid = self.post.pid
    post_shape = self.post.shape
    if isinstance(self.post, BaseNeuronSlice):
      post_slice = self.post.index
      if post_pid == mpi_rank:
        post = self.post.build()
      else:
        post = self.post.target
    elif isinstance(self.post, BaseNeuron):
      post_slice = None
      if post_pid == mpi_rank:
        post = self.post.build()
      else:
        post = self.post
    else:
      raise ValueError(type(self.post))

    if pre_pid == post_pid and pre_pid == mpi_rank:
      self.lowref = self.model_class(pre, post, self.conn, *self.args, **self.kwargs)
    elif pre_pid == mpi_rank:
      if post not in BaseNeuron.proxy_neurons:
        tmp_ = bpl.core.neurons.ProxyNeuronGroup(post_shape)
        BaseNeuron.proxy_neurons.update({post: tmp_})
      else:
        tmp_ = BaseNeuron.proxy_neurons[post]
      if post_slice is not None:
        for sli in post_slice:
          tmp_ = tmp_[sli]
      self.lowref = bpl.core.synapses.RemoteSynapse(
        synapse_class=self.model_class,
        param_dict=dict(pre=pre, post=tmp_, conn=self.conn, *self.args, **self.kwargs),
        source_rank=pre_pid,
        target_rank=post_pid)
    elif post_pid == mpi_rank:
      if pre not in BaseNeuron.proxy_neurons:
        tmp_ = bpl.core.neurons.ProxyNeuronGroup(pre_shape)
        BaseNeuron.proxy_neurons.update({pre: tmp_})
      else:
        tmp_ = BaseNeuron.proxy_neurons[pre]
      if pre_slice is not None:
        for sli in pre_slice:
          tmp_ = tmp_[sli]
      self.lowref = bpl.core.synapses.RemoteSynapse(
        synapse_class=self.model_class,
        param_dict=dict(pre=tmp_, post=post, conn=self.conn, *self.args, **self.kwargs),
        source_rank=pre_pid,
        target_rank=post_pid)
    return self.lowref


# another way to define respa LIF
# BaseNeuron.register(dyn.LIF)


class Network:

  def __init__(self, *ds_tuple, name: str = None, mode: Mode = normal, **ds_dict):
    self.ds_tuple = ds_tuple
    self.ds_dict = ds_dict
    self.lowref = bpl.core.Network((), name=name, mode=mode)

  def __getattr__(self, __name: str):
    return self.lowref.__getattribute__(__name)

  def build_all_population_synapse(self):
    for pop in ResManager.pops:
      pop.build()
    for syn in ResManager.syns:
      syn.build()
    self.lowref.register_implicit_nodes(*map(lambda x: x.lowref, ResManager.pops))
    self.lowref.register_implicit_nodes(*map(lambda x: x.lowref, ResManager.syns))

  def build(self):
    self.pops_ = []
    self.syns_ = []

    def reg_pop_syn(v):
      if isinstance(v, BaseNeuron):
        self.pops_.append(v)
      elif isinstance(v, BaseSynapse):
        self.syns_.append(v)

    jax.tree_util.tree_map(reg_pop_syn, self.__dict__)

    def simple_split(pops_):
      res = [[] for i in range(mpi_size)]
      avg = len(pops_) // mpi_size
      for i in range(mpi_size):
        res[i].extend(pops_[i * avg:i * avg + avg])
      res[-1].extend(pops_[avg * mpi_size:])
      return res

    self.pops_by_rank = simple_split(self.pops_)
    offset = 1
    for i in range(mpi_size):
      for __pops in self.pops_by_rank[i]:
        __pops.pid = i
    for node in self.pops_by_rank[mpi_rank]:
      node.neuron_start_id = offset
      offset += node.shape

      node.build()
      self.lowref.register_implicit_nodes(node.lowref)
    for node in self.syns_:
      if node.build() is not None:
        self.lowref.register_implicit_nodes(node.lowref)

  def update(self, *args, **kwargs):
    self.lowref.update(*args, **kwargs)

  def register_nodes(self, *nodes, **named_nodes):
    self.lowref.register_implicit_nodes(*nodes, **named_nodes)

  def register_vars(self, *variables, **named_variables):
    self.lowref.register_implicit_vars(*variables, **named_variables)


def input_transform(pops: Sequence):
  """Select BaseNeurons which locate on current process and return its input.

  Parameters
  ----------
  pops : Sequence

  Returns
  -------
  The inputs for the target DynamicalSystem. It should be the format
    of `[(target, value, [type, operation])]
  """
  if len(pops) > 0 and not isinstance(pops[0], (list, tuple)):
    pops = [pops]
  input_trans = []
  for pop in pops:
    try:
      tmp = pop[0].input
      input_trans.append((tmp, pop[1]) + pop[2:])
    except Exception:
      continue
  return input_trans


def monitor_transform(pops: Sequence[BaseNeuron], attr: str = 'spike'):
  """generate monitor parameters dictionary

  Parameters
  ----------
  pops : Sequence[BaseNeuron]

  attr : str, optional
      attribute to monitor, by default 'spike'

  Returns
  -------
  {attribute : BaseNeuron.attribute}
  """
  mon_var = {}
  for pop in pops:
    try:
      tmp = getattr(pop, attr)
      mon_var.update({attr: tmp})
    except Exception:
      continue
  return mon_var



from typing import Union, Dict, List, Tuple
import brainpy.math as bm
from brainpy.base.base import Base
import gc


class DynamicalSystem(Base):
  '''These variable is used in multi-device enviornment.'''
  remote_global_delay_data: Dict[str, Tuple[Union[bm.LengthDelay, None], bm.Variable]] = dict()

  remote_synapse_mark: List[str] = []

  def __del__(self):
    """Function for handling `del` behavior.

    This function is used to pop out the variables which registered in global delay data and remote_global_delay_data.
    """
    if hasattr(self, 'comm'):
      if hasattr(self, 'local_delay_vars'):
        for key in tuple(self.local_delay_vars.keys()):
          val = self.remote_global_delay_data.pop(key)
          del val
          val = self.local_delay_vars.pop(key)
          del val
    else:
      if hasattr(self, 'local_delay_vars'):
        for key in tuple(self.local_delay_vars.keys()):
          val = self.global_delay_data.pop(key)
          del val
          val = self.local_delay_vars.pop(key)
          del val
    if hasattr(self, 'implicit_nodes'):
      for key in tuple(self.implicit_nodes.keys()):
        del self.implicit_nodes[key]
    if hasattr(self, 'implicit_vars'):
      for key in tuple(self.implicit_vars.keys()):
        del self.implicit_vars[key]
    for key in tuple(self.__dict__.keys()):
      del self.__dict__[key]
    gc.collect()



import brainpy.dyn as dyn
from brainpy.connect import TwoEndConnector
from typing import Union

from .base import BaseSynapse, BaseNeuron, BaseNeuronSlice


class Exponential(BaseSynapse):

  def __init__(self,
               pre: Union[BaseNeuron, BaseNeuronSlice],
               post: Union[BaseNeuron, BaseNeuronSlice],
               conn: TwoEndConnector,
               g_max=1.0,
               tau=8.0,
               delay_step=None):
    kwargs = dict(g_max=g_max, tau=tau, delay_step=delay_step)
    super().__init__(pre, post, conn, **kwargs)
    self.model_class = dyn.synapses.Exponential
    # self.model_class_remote = bpl.synapses.RemoteExponential



from typing import Union, Callable, Optional

import brainpy.math as bm
from bpl.core.neurons_base import ProxyNeuGroup
from brainpy.initialize import (ZeroInit, Initializer, parameter, variable_)
from brainpy.modes import Mode, NormalMode, TrainingMode, normal, check_mode
from brainpy.types import Shape, Array


class ProxyNeuronGroup(ProxyNeuGroup):
  r"""Leaky integrate-and-fire neuron model in multi-device enviornment.
  """

  def __init__(
    self,
    size: Shape,
    keep_size: bool = False,

    # other parameter
    V_rest: Union[float, Array, Initializer, Callable] = 0.,
    V_reset: Union[float, Array, Initializer, Callable] = -5.,
    V_th: Union[float, Array, Initializer, Callable] = 20.,
    R: Union[float, Array, Initializer, Callable] = 1.,
    tau: Union[float, Array, Initializer, Callable] = 10.,
    tau_ref: Optional[Union[float, Array, Initializer, Callable]] = None,
    V_initializer: Union[Initializer, Callable, Array] = ZeroInit(),
    noise: Optional[Union[float, Array, Initializer, Callable]] = None,
    method: str = 'exp_auto',
    name: Optional[str] = None,

    # training parameter
    mode: Mode = normal,
    spike_fun: Callable = bm.spike_with_sigmoid_grad,
  ):
    # initialization
    super(ProxyNeuronGroup, self).__init__(size=size, name=name, keep_size=keep_size, mode=mode)
    check_mode(self.mode, (TrainingMode, NormalMode), self.__class__)

    # parameters
    # self.V_rest = parameter(V_rest, 0, allow_none=False)
    # self.V_reset = parameter(V_reset, 0, allow_none=False)
    # self.V_th = parameter(V_th, 0, allow_none=False)
    # self.tau = parameter(tau, 0, allow_none=False)
    # self.R = parameter(R, 0, allow_none=False)
    self.tau_ref = parameter(tau_ref, 0, allow_none=True)
    # self.noise = init_noise(noise, 0)
    # self.spike_fun = check_callable(spike_fun, 'spike_fun')

    # initializers
    # check_initializer(V_initializer, 'V_initializer')
    # self._V_initializer = V_initializer

    # variables
    # self.V = variable_(V_initializer, 0, mode)
    self.V = variable_(bm.zeros, 0, mode)
    self.input = variable_(bm.zeros, 0, mode)
    # Make sure the 'spike' attribute is consistent with real neuron group
    # It is useful for JIT in multi-device enviornment.
    sp_type = bm.dftype() if isinstance(mode, TrainingMode) else bool  # the gradient of spike is a float
    self.spike = variable_(lambda s: bm.zeros(s, dtype=sp_type), self.varshape, mode)

    if self.tau_ref is not None:
      self.t_last_spike = variable_(lambda s: bm.ones(s) * -1e7, 0, mode)
      self.refractory = variable_(lambda s: bm.zeros(s, dtype=bool), 0, mode)

  def update(self, tdi, x=None):
    pass



class ResManager():
  # [BaseNeuron(), ]
  pops = []

  # {id: BaseNeuron()}, id is BaseNeuron's id
  pops_by_id = {}

  # [BaseSynapse(), ]
  syns = []

  # element is an array, like [BaseNeuron(), ], pops_by_rank can be expressed as: {device_index: [BaseNeuron(), ] }
  pops_by_rank = {}

  @classmethod
  def clear(cls):
    cls.pops = []
    cls.pops_by_id = {}
    cls.syns = []
    cls.pops_by_rank = {}



from typing import Sequence
from .base import BaseNeuron


class Input():

  def __init__(self, pops: Sequence[BaseNeuron], value: float) -> None:
    self.pops = pops
    self.value = value


class MonitorKey():
  spike = 'spike'
  volt = "V"


class Monitor():

  def __init__(self, pops: Sequence[BaseNeuron], monitor_key: MonitorKey.spike) -> None:
    self.pops = pops
    self.monitor_key = monitor_key



from typing import Union, Sequence
import collections
import brainpy.math as bm
from brainpy import dyn
from brainpy.initialize import variable_
from brainpy.modes import Mode, TrainingMode, normal
from brainpy.types import Array, Shape


class ProxyNeuGroup(dyn.NeuGroup):
  """Base class to model neuronal groups in multi-device enviornment.
  """

  def __init__(
    self,
    size: Shape,
    keep_size: bool = False,
    name: str = None,
    mode: Mode = normal,
  ):
    # initialize
    super(ProxyNeuGroup, self).__init__(size=size, name=name, keep_size=keep_size, mode=mode)

  def __getitem__(self, item):
    return ProxyNeuGroupView(target=self, index=item, keep_size=self.keep_size)


class ProxyNeuGroupView(ProxyNeuGroup):
  """A view for a neuron group instance in multi-device enviornment."""

  def __init__(self,
               target: ProxyNeuGroup,
               index: Union[slice, Sequence, Array],
               name: str = None,
               mode: Mode = None,
               keep_size: bool = False):
    # check target
    if not isinstance(target, dyn.DynamicalSystem):
      raise TypeError(f'Should be instance of dyn.DynamicalSystem, but we got {type(target)}.')

    # check slicing
    if isinstance(index, (int, slice)):
      index = (index,)
    self.index = index  # the slice

    # check slicing
    var_shapes = target.varshape
    if len(self.index) > len(var_shapes):
      raise ValueError(f"Length of the index should be less than "
                       f"that of the target's varshape. But we "
                       f"got {len(self.index)} > {len(var_shapes)}")

    # get size
    size = []
    for i, idx in enumerate(self.index):
      if isinstance(idx, int):
        size.append(1)
      elif isinstance(idx, slice):
        size.append(dyn.base._slice_to_num(idx, var_shapes[i]))
      else:
        # should be a list/tuple/array of int
        # do not check again
        if not isinstance(idx, collections.Iterable):
          raise TypeError('Should be an iterable object of int.')
        size.append(len(idx))
    size += list(var_shapes[len(self.index):])

    # initialization
    ProxyNeuGroup.__init__(self, tuple(size), keep_size=keep_size, name=name, mode=mode)
    # Proxy neuron needs 'input' attribute during distributed network updating.
    self.input = variable_(bm.zeros, 0, mode)
    # Make sure the 'spike' attribute is consistent with real neuron group
    # It is useful for JIT in multi-device enviornment.
    sp_type = bm.dftype() if isinstance(mode, TrainingMode) else bool  # the gradient of spike is a float
    self.spike = variable_(lambda s: bm.zeros(s, dtype=sp_type), self.varshape, mode)
    # Proxy neuron needs 'V' attribute during distributed network updating.
    self.V = variable_(bm.zeros, 0, mode)



import brainpy.dyn as dyn
from typing import Union, Callable, List
import bpl
from .base import Network, input_transform, monitor_transform
from .device import Input, Monitor


class DSRunner:

  def __init__(
      self,
      target: Union[dyn.DynamicalSystem, Network],
      # inputs for target variables
      inputs: List[Input],
      monitors: List[Monitor],
      fun_inputs: Callable = None,
      # extra info
      dt: float = None,
      t0: Union[float, int] = 0.,
      spike_callback: Callable = None,
      volt_callback: Callable = None,
      **kwargs):
    if not isinstance(target, (Network, dyn.DynamicalSystem)):
      raise ValueError(type(target))

    inputs_l = []
    for input in inputs:
      inputs_l.append((input.pops, input.value))
    inputs_l = input_transform(inputs_l)
    monitors_d = {}
    for m in monitors:
      monitors_d.update(monitor_transform(m.pops, m.monitor_key))

    if isinstance(target, Network):
      target = target.lowref

    def _callback(t: float, d: dict):
      for k, v in d.items():
        if k == 'spike' and spike_callback:
          a = []
          for i, j in enumerate(v):
            if bool(j):
              a.append((i + 1, t))
          spike_callback(a)
        if k == 'V' and volt_callback:
          a = []
          for i, j in enumerate(v):
            a.append((i + 1, t, float(j)))
          volt_callback(a)

    c = _callback if spike_callback or volt_callback else None

    self.lowref = bpl.BplRunner(
      target=target, inputs=inputs_l, monitors=monitors_d, fun_inputs=fun_inputs, dt=dt, t0=t0, callback=c, **kwargs)

  def __getattr__(self, __name: str):
    return self.lowref.__getattribute__(__name)



from .base import BaseNeuron
import brainpy.dyn as dyn


class LIF(BaseNeuron):

  def __init__(self, size: int, V_reset: float, V_th: float, V_rest: float, tau: float, tau_ref: float):
    kwargs = dict(V_reset=V_reset, V_th=V_th, V_rest=V_rest, tau=tau, tau_ref=tau_ref)
    super(LIF, self).__init__(size, **kwargs)
    self.model_class = dyn.LIF



from typing import Union, Dict, Sequence
import brainpy.math as bm
from brainpy import dyn
from brainpy.modes import Mode, normal
from .base import DynamicalSystem
from mpi4py import MPI
import platform

if platform.system() != 'Windows':
  import mpi4jax
import numpy as np


class Network(dyn.Network, DynamicalSystem):
  """Exponential decay synapse model in multi-device environment.
  """

  def __init__(self, *ds_tuple, comm=MPI.COMM_WORLD, name: str = None, mode: Mode = normal, **ds_dict):
    super(Network, self).__init__(*ds_tuple, name=name, mode=mode, **ds_dict)
    self.comm = comm
    if self.comm is None:
      self.rank = None
    else:
      self.rank = self.comm.Get_rank()

  def update_local_delays(self, nodes: Union[Sequence, Dict] = None):
    """overwrite 'update_local_delays' method in brainpy's Network.
    """
    # update delays
    if nodes is None:
      nodes = tuple(self.nodes(level=1, include_self=False).subset(dyn.DynamicalSystem).unique().values())
    elif isinstance(nodes, dyn.DynamicalSystem):
      nodes = (nodes,)
    elif isinstance(nodes, dict):
      nodes = tuple(nodes.values())
    if not isinstance(nodes, (tuple, list)):
      raise ValueError('Please provide nodes as a list/tuple/dict of dyn.DynamicalSystem.')
    for node in nodes:
      if hasattr(node, 'comm'):
        for name in node.local_delay_vars:
          if self.rank == node.source_rank:
            if platform.system() == 'Windows':
              self.comm.send(len(node.synapse_instance.pre.spike), dest=node.target_rank, tag=2)
              self.comm.Send(node.synapse_instance.pre.spike.to_numpy(), dest=node.target_rank, tag=3)
            else:
              token = mpi4jax.send(node.synapse_instance.pre.spike.value, dest=node.target_rank, tag=3, comm=self.comm)
          elif self.rank == node.target_rank:
            delay = node.remote_global_delay_data[name][0]
            if platform.system() == 'Windows':
              pre_len = self.comm.recv(source=node.source_rank, tag=2)
              target = np.empty(pre_len, dtype=np.bool_)
              self.comm.Recv(target, source=node.source_rank, tag=3)
            else:
              target, token = mpi4jax.recv(
                node.synapse_instance.pre.spike.value, source=node.source_rank, tag=3, comm=self.comm)
            target = bm.Variable(target)
            delay.update(target.value)
      else:
        for name in node.local_delay_vars:
          delay = self.global_delay_data[name][0]
          target = self.global_delay_data[name][1]
          delay.update(target.value)

  def reset_local_delays(self, nodes: Union[Sequence, Dict] = None):
    """overwrite 'reset_local_delays' method in brainpy's Network.
    """
    # reset delays
    if nodes is None:
      nodes = self.nodes(level=1, include_self=False).subset(dyn.DynamicalSystem).unique().values()
    elif isinstance(nodes, dict):
      nodes = nodes.values()
    for node in nodes:
      if hasattr(node, 'comm'):
        for name in node.local_delay_vars:
          if self.rank == node.source_rank:
            if platform.system() == 'Windows':
              self.comm.send(len(node.synapse_instance.pre.spike), dest=node.target_rank, tag=4)
              self.comm.Send(node.synapse_instance.pre.spike.to_numpy(), dest=node.target_rank, tag=5)
            else:
              token = mpi4jax.send(node.synapse_instance.pre.spike.value, dest=node.target_rank, tag=4, comm=self.comm)
          elif self.rank == node.target_rank:
            delay = node.remote_global_delay_data[name][0]
            if platform.system() == 'Windows':
              pre_len = self.comm.recv(source=node.source_rank, tag=4)
              target = np.empty(pre_len, dtype=np.bool_)
              self.comm.Recv(target, source=node.source_rank, tag=5)
            else:
              target, token = mpi4jax.recv(
                node.synapse_instance.pre.spike.value, source=node.source_rank, tag=4, comm=self.comm)
            target = bm.Variable(target)
            delay.reset(target.value)
      else:
        for name in node.local_delay_vars:
          delay = self.global_delay_data[name][0]
          target = self.global_delay_data[name][1]
          delay.reset(target.value)
